from typing import Any, Callable

from pydantic_core import PydanticCustomError

# NOTE: We use the factory approach because PydanticCustomError is a final class.
#   That is also why this module is internal.


def CustomError(fn: Callable, invalid_tag: str, **kwargs) -> PydanticCustomError:
    return PydanticCustomError(fn.__name__, f"Invalid {invalid_tag}", kwargs)


def HexValueError(value: Any) -> PydanticCustomError:
    return CustomError(HexValueError, "hex value", value=value)


def SizeError(size: Any, value: Any) -> PydanticCustomError:
    return CustomError(SizeError, "size of value", size=size, value=value)


def Bip122UriFormatError(value: str) -> PydanticCustomError:
    return CustomError(
        Bip122UriFormatError,
        "BIP-122 URI format",
        uri=value,
        format="blockchain://<genesis_hash>/block/<block_hash>.",
    )



from typing import Any, ClassVar, Optional, Tuple

from eth_utils import is_checksum_address, to_checksum_address
from pydantic_core import CoreSchema
from pydantic_core.core_schema import (
    ValidationInfo,
    str_schema,
    with_info_before_validator_function,
)

from eth_pydantic_types.hex import BaseHexStr, HexBytes
from eth_pydantic_types.validators import validate_address_size

ADDRESS_PATTERN = "^0x[a-fA-F0-9]{40}$"


def address_schema():
    return str_schema(min_length=42, max_length=42, pattern=ADDRESS_PATTERN)


class Address(BaseHexStr):
    """
    Use for address-types. Validates as a checksummed address. Left-pads zeroes
    if necessary.
    """

    schema_pattern: ClassVar[str] = ADDRESS_PATTERN
    schema_examples: ClassVar[Tuple[str, ...]] = (
        "0x0000000000000000000000000000000000000000",  # Zero address
        "0x02c84e944F97F4A4f60221e6fb5d5DbAE49c7aaB",  # Leading zero
        "0xa5a13f62ce1113838e0d9b4559b8caf5f76463c0",  # Trailing zero
        "0x1e59ce931B4CFea3fe4B875411e280e173cB7A9C",
    )

    def __get_pydantic_core_schema__(self, *args, **kwargs) -> CoreSchema:
        schema = with_info_before_validator_function(
            self._validate_address,
            address_schema(),
        )
        return schema

    @classmethod
    def _validate_address(cls, value: Any, info: Optional[ValidationInfo] = None) -> str:
        if isinstance(value, str) and is_checksum_address(value):
            return value

        elif not isinstance(value, str):
            value = HexBytes(value).hex()

        number = value[2:] if value.startswith("0x") else value
        number_padded = validate_address_size(number)
        value = f"0x{number_padded}"
        return to_checksum_address(value)



from enum import Enum
from functools import cached_property
from typing import Any, Optional, Tuple

from pydantic_core import CoreSchema
from pydantic_core.core_schema import (
    ValidationInfo,
    str_schema,
    with_info_before_validator_function,
)

from eth_pydantic_types._error import Bip122UriFormatError
from eth_pydantic_types.hex import validate_hex_str


class Bip122UriType(Enum):
    TX = "tx"
    BLOCK = "block"
    ADDRESS = "address"


class Bip122Uri(str):
    prefix: str = "blockchain://"

    @classmethod
    def __get_pydantic_json_schema__(cls, core_schema, handler):
        json_schema = handler(core_schema)
        example = (
            f"{cls.prefix}d4e56740f876aef8c010b86a40d5f56745a118d0906a34e69aec8c0db1cb8fa3"
            f"/{Bip122UriType.BLOCK.value}/"
            f"752820c0ad7abc1200f9ad42c4adc6fbb4bd44b5bed4667990e64565102c1ba6"
        )
        pattern = f"^{cls.prefix}[0-9a-f]{{64}}/{Bip122UriType.BLOCK.value}/[0-9a-f]{{64}}$"
        json_schema.update(examples=[example], pattern=pattern)
        return json_schema

    def __get_pydantic_core_schema__(self, *args, **kwargs) -> CoreSchema:
        return with_info_before_validator_function(
            self._validate,
            str_schema(),
        )

    @classmethod
    def _validate(cls, value: Any, info: Optional[ValidationInfo] = None) -> str:
        if not value.startswith(cls.prefix):
            raise Bip122UriFormatError(value)

        genesis_hash, block_keyword, block_hash = cls.parse(value)
        return f"{cls.prefix}{genesis_hash[2:]}/{block_keyword.value}/{block_hash[2:]}"

    @classmethod
    def parse(cls, value: str) -> Tuple[str, Bip122UriType, str]:
        protocol_suffix = value.replace(cls.prefix, "")
        protocol_parsed = protocol_suffix.split("/")
        if len(protocol_parsed) != 3:
            raise Bip122UriFormatError(value)

        genesis_hash, block_keyword, block_hash = protocol_parsed
        block_keyword = block_keyword.lower()
        if block_keyword not in [x.value for x in Bip122UriType]:
            raise Bip122UriFormatError(value)

        return (
            validate_hex_str(genesis_hash),
            Bip122UriType(block_keyword),
            validate_hex_str(block_hash),
        )

    @cached_property
    def parsed(self) -> Tuple[str, Bip122UriType, str]:
        return self.parse(self)

    @property
    def chain(self) -> str:
        return self.parsed[0]

    @property
    def uri_type(self) -> Bip122UriType:
        return self.parsed[1]

    @property
    def hash(self) -> str:
        return self.parsed[2]



from typing import Any, ClassVar, Optional, Tuple, Union

from hexbytes import HexBytes as BaseHexBytes
from pydantic_core import CoreSchema
from pydantic_core.core_schema import (
    ValidationInfo,
    bytes_schema,
    no_info_before_validator_function,
    str_schema,
    with_info_before_validator_function,
)

from eth_pydantic_types._error import HexValueError
from eth_pydantic_types.serializers import hex_serializer

schema_pattern = "^0x([0-9a-f][0-9a-f])*$"
schema_examples = (
    "0x",  # empty bytes
    "0xd4",
    "0xd4e5",
    "0xd4e56740",
    "0xd4e56740f876aef8",
    "0xd4e56740f876aef8c010b86a40d5f567",
    "0xd4e56740f876aef8c010b86a40d5f56745a118d0906a34e69aec8c0db1cb8fa3",
)


class BaseHex:
    schema_pattern: ClassVar[str] = schema_pattern
    schema_examples: ClassVar[Tuple[str, ...]] = schema_examples

    @classmethod
    def __get_pydantic_json_schema__(cls, core_schema, handler):
        json_schema = handler(core_schema)
        json_schema.update(
            format="binary", pattern=cls.schema_pattern, examples=list(cls.schema_examples)
        )
        return json_schema


class HexBytes(BaseHexBytes, BaseHex):
    """
    Use when receiving ``hexbytes.HexBytes`` values. Includes
    a pydantic validator and serializer.
    """

    def __get_pydantic_core_schema__(self, *args, **kwargs) -> CoreSchema:
        schema = with_info_before_validator_function(self._validate_hexbytes, bytes_schema())
        schema["serialization"] = hex_serializer
        return schema

    @classmethod
    def fromhex(cls, hex_str: str) -> "HexBytes":
        value = hex_str[2:] if hex_str.startswith("0x") else hex_str
        return super().fromhex(value)

    @classmethod
    def _validate_hexbytes(cls, value: Any, info: Optional[ValidationInfo] = None) -> BaseHexBytes:
        return BaseHexBytes(value)


class BaseHexStr(str, BaseHex):
    @classmethod
    def from_bytes(cls, data: bytes) -> "BaseHexStr":
        hex_str = data.hex()
        return cls(hex_str if hex_str.startswith("0x") else hex_str)

    @classmethod
    def validate_hex(cls, data: Union[bytes, str, int]):
        if isinstance(data, bytes):
            return cls.from_bytes(data)

        elif isinstance(data, str):
            return validate_hex_str(data)

        elif isinstance(data, int):
            return BaseHexBytes(data).hex()

        raise HexValueError(data)

    def __int__(self) -> int:
        return int(self, 16)

    def __bytes__(self) -> bytes:
        return bytes.fromhex(self[2:])


class HexStr(BaseHexStr):
    """A hex string value, typically from a hash."""

    def __get_pydantic_core_schema__(cls, *args, **kwargs):
        return no_info_before_validator_function(cls.validate_hex, str_schema())

    @classmethod
    def from_bytes(cls, data: bytes) -> "HexStr":
        return HexStr(super().from_bytes(data))


def validate_hex_str(value: str) -> str:
    hex_value = (value[2:] if value.startswith("0x") else value).lower()
    if set(hex_value) - set("1234567890abcdef"):
        raise HexValueError(value)

    # Missing zero padding.
    if len(hex_value) % 2 != 0:
        hex_value = f"0{hex_value}"

    return f"0x{hex_value}"



import pytest
from hexbytes import HexBytes as BaseHexBytes
from pydantic import BaseModel, ValidationError

from eth_pydantic_types.hex import HexBytes, HexStr


class BytesModel(BaseModel):
    value: HexBytes


class StrModel(BaseModel):
    value: HexStr


@pytest.mark.parametrize("value", ("0xa", 10, b"\n"))
def test_hexbytes(value):
    actual = BytesModel(value=value)

    # The end result, the value is a hexbytes.HexBytes
    assert actual.value == BaseHexBytes(value)
    assert actual.value.hex() == "0x0a"
    assert isinstance(actual.value, bytes)
    assert isinstance(actual.value, BaseHexBytes)


def test_invalid_hexbytes():
    with pytest.raises(ValidationError):
        BytesModel(value="foo")


def test_hexbytes_fromhex(bytes32str):
    actual_with_0x = HexBytes.fromhex(bytes32str)
    actual_without_0x = HexBytes.fromhex(bytes32str[2:])
    expected = HexBytes(bytes32str)
    assert actual_with_0x == actual_without_0x == expected


def test_hexbytes_schema():
    actual = BytesModel.model_json_schema()
    prop = actual["properties"]["value"]
    assert prop["type"] == "string"
    assert prop["format"] == "binary"
    assert prop["pattern"] == "^0x([0-9a-f][0-9a-f])*$"
    assert prop["examples"] == [
        "0x",
        "0xd4",
        "0xd4e5",
        "0xd4e56740",
        "0xd4e56740f876aef8",
        "0xd4e56740f876aef8c010b86a40d5f567",
        "0xd4e56740f876aef8c010b86a40d5f56745a118d0906a34e69aec8c0db1cb8fa3",
    ]


def test_hexbytes_model_dump(bytes32str):
    model = BytesModel(value=bytes32str)
    actual = model.model_dump()
    expected = {"value": "0x9b70bd98ccb5b6434c2ead14d68d15f392435a06ff469f8d1f8cf38b2ae0b0e2"}
    assert actual == expected


@pytest.mark.parametrize("value", ("0xa", 10, HexBytes(10)))
def test_hexstr(value):
    actual = StrModel(value=value)

    # The end result, the value is a str
    assert actual.value == "0x0a"
    assert isinstance(actual.value, str)


def test_invalid_hexstr():
    with pytest.raises(ValidationError):
        StrModel(value="foo")


def test_hexstr_conversions():
    model = StrModel(value="0x123")
    assert int(model.value, 16) == 291
    assert bytes.fromhex(model.value[2:]) == b"\x01#"


def test_hexstr_schema():
    actual = StrModel.model_json_schema()
    properties = actual["properties"]
    assert len(properties) == 1
    prop = properties["value"]
    assert prop["type"] == "string"
    assert prop["format"] == "binary"
    assert prop["pattern"] == "^0x([0-9a-f][0-9a-f])*$"
    assert prop["examples"] == [
        "0x",
        "0xd4",
        "0xd4e5",
        "0xd4e56740",
        "0xd4e56740f876aef8",
        "0xd4e56740f876aef8c010b86a40d5f567",
        "0xd4e56740f876aef8c010b86a40d5f56745a118d0906a34e69aec8c0db1cb8fa3",
    ]


def test_hexstr_model_dump(bytes32str):
    model = StrModel(value=bytes32str)
    actual = model.model_dump()
    expected = {"value": "0x9b70bd98ccb5b6434c2ead14d68d15f392435a06ff469f8d1f8cf38b2ae0b0e2"}
    assert actual == expected

    model = StrModel(value=3)
    actual = model.model_dump()
    expected = {"value": "0x03"}
    assert actual == expected



from typing import Any, ClassVar, Optional, Tuple, Type

from pydantic_core.core_schema import (
    CoreSchema,
    ValidationInfo,
    bytes_schema,
    str_schema,
    with_info_before_validator_function,
)

from eth_pydantic_types.hex import BaseHexStr, HexBytes
from eth_pydantic_types.serializers import hex_serializer
from eth_pydantic_types.validators import validate_bytes_size, validate_str_size


def _get_hash_pattern(str_size: int) -> str:
    return f"^0x[a-fA-F0-9]{{{str_size}}}$"


def _get_hash_examples(str_size: int) -> Tuple[str, str, str, str]:
    zero_hash = f"0x{'0' * str_size}"
    leading_zero = f"0x01{'1e' * ((str_size - 1) // 2)}"
    trailing_zero = f"0x{'1e' * ((str_size - 1) // 2)}10"
    full_hash = f"0x{'1e' * (str_size // 2)}"
    return zero_hash, leading_zero, trailing_zero, full_hash


class HashBytes(HexBytes):
    """
    Represents a single-slot static hash as bytes.
    This type is meant to be overridden by the larger hash types with a new size.
    The class variable "size" is overridden in subclasses for each byte-size,
    e.g. HashBytes20, HashBytes32.
    """

    size: ClassVar[int] = 1
    schema_pattern: ClassVar[str] = _get_hash_pattern(1)
    schema_examples: ClassVar[Tuple[str, ...]] = _get_hash_examples(1)

    def __get_pydantic_core_schema__(self, *args, **kwargs) -> CoreSchema:
        schema = with_info_before_validator_function(
            self._validate_hash, bytes_schema(max_length=self.size, min_length=self.size)
        )
        schema["serialization"] = hex_serializer
        return schema

    @classmethod
    def _validate_hash(cls, value: Any, info: Optional[ValidationInfo] = None) -> bytes:
        return cls(cls.validate_size(HexBytes(value)))

    @classmethod
    def validate_size(cls, value: bytes) -> bytes:
        return validate_bytes_size(value, cls.size)


class HashStr(BaseHexStr):
    """
    Represents a single-slot static hash as a str.
    This type is meant to be overridden by the larger hash types with a new size.
    e.g. HashStr20, HashStr32.
    """

    size: ClassVar[int] = 1
    schema_pattern: ClassVar[str] = _get_hash_pattern(1)
    schema_examples: ClassVar[Tuple[str, ...]] = _get_hash_examples(1)

    def __get_pydantic_core_schema__(self, *args, **kwargs) -> CoreSchema:
        str_size = self.size * 2 + 2
        return with_info_before_validator_function(
            self._validate_hash, str_schema(max_length=str_size, min_length=str_size)
        )

    @classmethod
    def _validate_hash(cls, value: Any, info: Optional[ValidationInfo] = None) -> str:
        hex_str = cls.validate_hex(value)
        hex_value = hex_str[2:] if hex_str.startswith("0x") else hex_str
        sized_value = cls.validate_size(hex_value)
        return cls(f"0x{sized_value}")

    @classmethod
    def validate_size(cls, value: str) -> str:
        return validate_str_size(value, cls.size * 2)


def _make_hash_cls(size: int, base_type: Type):
    if issubclass(base_type, bytes):
        suffix = "Bytes"
        base_type = HashBytes
    else:
        suffix = "Str"
        base_type = HashStr

    str_size = size * 2
    return type(
        f"Hash{suffix}{size}",
        (base_type,),
        dict(
            size=size,
            schema_pattern=_get_hash_pattern(str_size),
            schema_examples=_get_hash_examples(str_size),
        ),
    )


HashBytes4 = _make_hash_cls(4, bytes)
HashBytes8 = _make_hash_cls(8, bytes)
HashBytes16 = _make_hash_cls(16, bytes)
HashBytes20 = _make_hash_cls(20, bytes)
HashBytes32 = _make_hash_cls(32, bytes)
HashBytes64 = _make_hash_cls(64, bytes)
HashStr4 = _make_hash_cls(4, str)
HashStr8 = _make_hash_cls(8, str)
HashStr16 = _make_hash_cls(16, str)
HashStr20 = _make_hash_cls(20, str)
HashStr32 = _make_hash_cls(32, str)
HashStr64 = _make_hash_cls(64, str)



from typing import Any, Callable, Dict, Optional, Sized, TypeVar, cast

from pydantic import WithJsonSchema
from pydantic_core.core_schema import bytes_schema

from eth_pydantic_types._error import SizeError

__SIZED_T = TypeVar("__SIZED_T", bound=Sized)


class WithBytesSchema(WithJsonSchema):
    def __init__(self, **kwargs):
        mode = kwargs.pop("mode", None)
        schema = cast(Dict[str, Any], bytes_schema(**kwargs))
        super().__init__(schema, mode=mode)


def validate_size(value: __SIZED_T, size: int, coerce: Optional[Callable] = None) -> __SIZED_T:
    if len(value) == size:
        return value

    elif coerce:
        return validate_size(coerce(value), size)

    raise SizeError(size, value)


def validate_bytes_size(value: bytes, size: int) -> bytes:
    return validate_size(value, size, coerce=lambda v: _left_pad_bytes(v, size))


def validate_address_size(value: str) -> str:
    return validate_str_size(value, 40)


def validate_str_size(value: str, size: int) -> str:
    return validate_size(value, size, coerce=lambda v: _left_pad_str(v, size))


def _left_pad_str(val: str, length: int) -> str:
    return "0" * (length - len(val)) + val if len(val) < length else val


def _left_pad_bytes(val: bytes, num_bytes: int) -> bytes:
    return b"\x00" * (num_bytes - len(val)) + val if len(val) < num_bytes else val



from .address import Address
from .bip122 import Bip122Uri
from .hash import (
    HashBytes4,
    HashBytes8,
    HashBytes16,
    HashBytes20,
    HashBytes32,
    HashBytes64,
    HashStr4,
    HashStr8,
    HashStr16,
    HashStr20,
    HashStr32,
    HashStr64,
)
from .hex import HexBytes, HexStr

__all__ = [
    "Address",
    "Bip122Uri",
    "HashBytes4",
    "HashBytes8",
    "HashBytes16",
    "HashBytes20",
    "HashBytes32",
    "HashBytes64",
    "HashStr4",
    "HashStr8",
    "HashStr16",
    "HashStr20",
    "HashStr32",
    "HashStr64",
    "HexBytes",
    "HexStr",
]



import pytest
from pydantic import BaseModel, ValidationError

from eth_pydantic_types.hash import (
    HashBytes8,
    HashBytes16,
    HashBytes32,
    HashBytes64,
    HashStr8,
    HashStr16,
    HashStr32,
    HashStr64,
)
from eth_pydantic_types.hex import HexBytes


class Model(BaseModel):
    valuebytes8: HashBytes8
    valuebytes16: HashBytes16
    valuebytes32: HashBytes32
    valuebytes64: HashBytes64
    valuestr8: HashStr8
    valuestr16: HashStr16
    valuestr32: HashStr32
    valuestr64: HashStr64

    @classmethod
    def from_single(cls, value):
        return cls(
            valuebytes8=value,
            valuebytes16=value,
            valuebytes32=value,
            valuebytes64=value,
            valuestr8=value,
            valuestr16=value,
            valuestr32=value,
            valuestr64=value,
        )


def test_hashbytes_fromhex(bytes32str):
    actual_with_0x = HashBytes32.fromhex(bytes32str)
    actual_without_0x = HashBytes32.fromhex(bytes32str[2:])
    expected = HexBytes(bytes32str)
    assert actual_with_0x == actual_without_0x == expected


def test_hashbytes_is_bytes(bytes32str):
    assert isinstance(HashBytes32.fromhex(bytes32str), bytes)


@pytest.mark.parametrize("value", ("0x32", HexBytes("0x32"), b"2", 50))
def test_hash(value):
    model = Model.from_single(value)
    assert len(model.valuebytes8) == 8
    assert len(model.valuebytes16) == 16
    assert len(model.valuebytes32) == 32
    assert len(model.valuebytes64) == 64
    assert len(model.valuestr8) == 18
    assert len(model.valuestr16) == 34
    assert len(model.valuestr32) == 66
    assert len(model.valuestr64) == 130
    assert model.valuebytes8.hex().endswith("32")
    assert model.valuebytes16.hex().endswith("32")
    assert model.valuebytes32.hex().endswith("32")
    assert model.valuestr64.endswith("32")
    assert model.valuestr8.endswith("32")
    assert model.valuestr16.endswith("32")
    assert model.valuestr32.endswith("32")
    assert model.valuestr64.endswith("32")


@pytest.mark.parametrize("value", ("foo", -35, "0x" + ("F" * 100)))
def test_invalid_hash(value):
    with pytest.raises(ValidationError):
        Model.from_single(value)


def test_schema():
    actual = Model.model_json_schema()
    for name, prop in actual["properties"].items():
        is_bytes = "bytes" in name
        if is_bytes:
            size_from_name = int(name.replace("valuebytes", ""))
            expected_length = size_from_name
        else:
            size_from_name = int(name.replace("valuestr", ""))
            expected_length = size_from_name * 2 + 2

        hex_value_str_size = size_from_name * 2

        assert prop["maxLength"] == expected_length
        assert prop["minLength"] == expected_length
        assert prop["type"] == "string"
        assert prop["format"] == "binary"
        assert prop["pattern"] == f"^0x[a-fA-F0-9]{{{hex_value_str_size}}}$"

        assert prop["examples"][0] == f"0x{'0' * hex_value_str_size}"
        assert prop["examples"][1].startswith("0x0")  # Leading zero
        assert prop["examples"][2].endswith("0")  # Trailing zero
        assert all(len(ex) == hex_value_str_size + 2 for ex in prop["examples"])


def test_model_dump(bytes32str):
    model = Model.from_single(5)
    actual = model.model_dump()
    expected = {
        "valuebytes8": "0x0000000000000005",
        "valuestr8": "0x0000000000000005",
        "valuebytes16": "0x00000000000000000000000000000005",
        "valuestr16": "0x00000000000000000000000000000005",
        "valuebytes32": "0x0000000000000000000000000000000000000000000000000000000000000005",
        "valuestr32": "0x0000000000000000000000000000000000000000000000000000000000000005",
        "valuebytes64": "0x00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000005",  # noqa: E501
        "valuestr64": "0x00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000005",  # noqa: E501
    }
    assert actual == expected
